(in-package :cl)
(defpackage :magrathea.file-manager
  (:use :clim :clim-lisp)
  (:import-from :uiop
   :directory-files
   :enough-pathname
   :pathname-parent-directory-pathname
   :pathname-root
   :directory-exists-p
   :subdirectories)
  (:export :run-file-manager))
(in-package :magrathea.file-manager)

(defclass file ()
  ((path :initarg :path :accessor %path)))
(defclass dir (file) nil)
(defun make-dir (path)
  (make-instance 'dir :path path))
(defmethod file-path ((obj file))
  (%path obj))

(define-application-frame file-manager ()
  ((directory :initform *default-pathname-defaults*
              :initarg :directory
              :accessor %directory))
  (:pointer-documentation t)
  (:panes (app :application
               :height 300 :width 600
               :display-function 'display-files)
          (int :interactor :height '(fill :relative)
                           :width '(fill :relative)))
  (:layouts (default (vertically () app int))))

(defun current-directory ()
  (%directory *application-frame*))

(defmethod change-directory ((frame file-manager) (dir dir))
  (setf (%directory frame) (%path dir)))

(defun display-files (frame pane)
  (let* ((dir     (current-directory))
         (files   (directory-files dir)) ;; XXX - `files' are not instances of `file'
         (parent  (make-dir (pathname-parent-directory-pathname dir)))
         (subdir-paths (subdirectories dir))
         (subdirs (mapcar #'make-dir subdir-paths)))
    ;; (format *debug-io* "display-files: subdirs -~%~a~%files -~%~a~%" subdir-paths files)
    (unless (equal dir (pathname-root dir))
      (with-output-as-presentation (pane parent 'dir)
        (format pane "↩~%")))
    (loop for subdir in subdirs do
      (with-output-as-presentation (pane subdir 'dir)
        (format pane "~a~%" (enough-pathname (file-path subdir) dir))))
    (loop for file in files do
      (with-output-as-presentation (pane file 'file)
        (format pane "~a~%" (enough-pathname file dir))))))

(define-file-manager-command (com-visit :name t) ((directory 'dir :gesture :select))
  (change-directory *application-frame* directory))

(define-file-manager-command (com-quit :name t) ()
  (frame-exit *application-frame*))

(defun run-file-manager (&optional dir)
  (run-frame-top-level (make-application-frame 'file-manager :directory dir))
  (resize-sheet (frame-top-level-sheet *application-frame*) 'fill 'fill))
